#include <cairo.h>
#include <gtk/gtk.h>
#include <math.h>
#include <unistd.h>

#include "support.h"
#include "interface.h"
#include "main.h"
#include "pref.h"
#include "gui.h"
#include "plot.h"

#define DANGLE	1.0		// Make 1 degree steps
#define IOD     200.0		// Inter-ocular distance
#define fovy    44.0		//field of view in y-axis
#define nearZ   100.0		//near clipping plane
#define farZ    5000.0		//far clipping plane
#define screenZ 1300.0		//screen projection plane
#define camZ    1500.0


// #define ST_LOOKAT
// #define ST_COLORMASK
#define ST_ACCUM		// Sum colors by using the accumulator

enum {	XY_plane, 
	ZX_plane, 
	YZ_plane 
};

enum {  LEFT_EYE,
	RIGHT_EYE
};

struct {
  char *mark[4];
} axismarkers[] = {
  {"X", "Y", "-X", "-Y"},
  {"X", "Z", "-X", "-Z"},
  {"Z", "Y", "-Z", "-Y"}
};

GLfloat LightPos[]		= {0.0, 1000.0, 1000.0, 1.0};
GLfloat AmbientMat[]		= {1.0, 1.0, 1.0, 1.0};
GLfloat DiffuseMat[]		= {1.8, 1.8, 0.0, 1.0}; 
GLfloat SpecularMat[]		= {0.1, 0.1, 0.1, 1.0};
GLfloat SpecularReflection[]    = {0.1, 0.1, 0.1, 1.0};

GLfloat AmbientMat_L[]		= {1.0, 0.0, 0.0, 1.0};       // Red
//GLfloat AmbientMat_L[]		= {0.0, 0.3, 0.7, 1.0};       // Cyan
GLfloat AmbientMat_R[]		= {0.0, 0.3, 1.0, 1.0};       // Cyan

GLfloat Black[]			= {0.0, 0.0, 0.0, 1.0};
GLfloat White[]			= {1.0, 1.0, 1.0, 1.0};
  
typedef struct {
  real radius;
  char *label;
} gain_circle;

typedef struct {
  char *name, 
       *lbl_N, *lbl_E, *lbl_W, *lbl_S;
  gain_circle circle[];
} plot_info;

plot_info logv_plot = {
  "log(V)",  "-Y", "-X", "+X", "+Y",
        {{ 0.0,	"-50dB"},        { 0.2,	"-40dB"},
         { 0.4,	"-30dB"},        { 0.6,	"-20dB"},
         { 0.8,	"-10dB"},        { 1.0,	"  0dB"},
         {-1.0}}
};

plot_info logp_plot = {
  "log(P)",  "-Y", "-X", "+X", "+Y",
        {{ 0.0,	"-50dB"},        { 0.2,	"-40dB"},
         { 0.4,	"-30dB"},        { 0.6,	"-20dB"},
         { 0.8,	"-10dB"},        { 1.0,	"  0dB"},
         {-1.0}}
};

plot_info lin_plot = {
  "lin",  "-Y", "-X", "+X", "+Y",
        {{ 0.0,	"0.0"},         { 0.2,	"0.2"},
         { 0.4,	"0.4"},         { 0.6,	"0.6"},
         { 0.8,	"0.8"},         { 1.0,	"1.0"},
         {-1.0}}
};	

plot_info arrl_plot = {
  "ARRL",  "-Y", "-X", "+X", "+Y",
        {{ 0.0303,      "-60dB"},       { 0.0972,       "-40dB"},
         { 0.1740,	"-30dB"},       { 0.3118,	"-20dB"},
         { 0.4172,	"-15dB"},       { 0.5584,	"-10dB"},
         { 0.7050,      " -6dB"},       { 0.8396,	" -3dB"},
         { 1.0,	        "  0dB"},
         {-1.0}}
};

int attrlist[] = {
    GDK_GL_RGBA,
    GDK_GL_RED_SIZE,		1,
    GDK_GL_GREEN_SIZE,		1,
    GDK_GL_BLUE_SIZE,		1,
    GDK_GL_DOUBLEBUFFER,
    GDK_GL_NONE
};

GdkGLContext    *threed_glcontext = NULL;
GdkGLDrawable   *threed_gldrawable = NULL;
GtkWidget       *threed_darea = NULL;

GdkGLConfig     *stereo_3D_glconfig = NULL;
GdkGLContext    *stereo_glcontext = NULL;
GdkGLDrawable   *stereo_gldrawable = NULL;
GtkWidget       *stereo_darea = NULL;

real	aspect = 1.0;
real	sphi   = 0.0,
	stheta = 0.0,
	sscale = 1.0,
	beginX = 0,
	beginY = 0;
	
GLuint  listID;

void    draw_3D_axis(void);
void    draw_3D_plot(GLfloat Amb[], GLfloat Dif[]);
void    draw_stereo_axis(GLfloat Amb[]);
void    draw_stereo_plot(GLfloat Amb[]);


real
to_rad(real angle)
{
  return angle*M_PI/180.0;
}

real
to_deg(real angle)
{
  return angle/M_PI*180.0;
}

complex
to_rect(real magn, real phase)
{
  return magn * cos(phase) + I * magn * sin(phase);
}

real
sum_to_radius(real sum)
{
  real res = sum;
  switch (pref.plot_scale) {
    case PLOT_SCALE_LIN:
      break;

    case PLOT_SCALE_LOGV:
      res = 1.0 + 20.0*log10(res)/50.0;
      break;

    case PLOT_SCALE_LOGP:
      res = 1.0 + 2*20.0*log10(res)/50.0;
      break;

    case PLOT_SCALE_ARRL:
      res = pow(0.89, -10.0*log10(res));
      break;
  }
  if (res < 0) res = 0.0;

  return res * PLOT_RADIUS;
}

void
draw_2D_axis(cairo_t *cr, int dir)
{
  int i;
  char atext[10];
  real xc, yc, dA,
    endx, endy,
    dashes[] = {2.0};
  plot_info *info;
  cairo_text_extents_t extents;

  xc = PLOT_WIDTH/2;
  yc = PLOT_HEIGHT/2;
  dA = 2. * M_PI / PLOT_NR_ANGLES;
  switch (pref.plot_scale) {
    case PLOT_SCALE_LIN:
      info = &lin_plot;
      break;
    case PLOT_SCALE_LOGV:
      info = &logv_plot;
      break;
    case PLOT_SCALE_LOGP:
      info = &logp_plot;
      break;
    case PLOT_SCALE_ARRL:
      info = &arrl_plot;
      break;
  }

  cairo_set_line_width(cr, 1.0);
  cairo_set_source_rgb(cr, PLOT_AXIS_R, PLOT_AXIS_G, PLOT_AXIS_B);

  // Angles
  cairo_select_font_face(cr, PLOT_AXIS_FONT, CAIRO_FONT_SLANT_NORMAL,
                               CAIRO_FONT_WEIGHT_BOLD);
  cairo_set_font_size(cr, PLOT_AXIS_SIZE);


  for (i = 0; i < PLOT_NR_ANGLES; i++) {
    cairo_move_to(cr, xc, yc);
    endx = xc + cos(i * dA) * PLOT_RADIUS;
    endy = yc + sin(i * dA) * PLOT_RADIUS;
    cairo_line_to(cr, endx, endy);
    cairo_stroke(cr);

    cairo_save(cr);
    cairo_move_to(cr, endx, endy);
    sprintf(atext, "%-3.0f", i*dA*180/M_PI);
    cairo_rotate(cr, i*dA + M_PI);
    cairo_show_text(cr, atext);
    cairo_restore(cr);
  }

  // Gain circles
  cairo_set_dash(cr, NULL, 0, 0.);
  for (i = 0; info->circle[i].radius >= 0.; i++) {
    cairo_new_sub_path(cr);
    cairo_arc(cr, xc, yc, info->circle[i].radius * PLOT_RADIUS, 0.0, 2.0*M_PI);
    cairo_stroke(cr);
  }

  // Show labels for the axis
  #define MARKER_RADIUS (0.4 * PLOT_WIDTH)
  cairo_select_font_face (cr, "Sans",
	CAIRO_FONT_SLANT_NORMAL,
        CAIRO_FONT_WEIGHT_NORMAL);
  cairo_set_font_size (cr, 50.0);
  for (i = 0; i < 3; i++) {
    cairo_text_extents (cr, axismarkers[dir].mark[i], &extents);
    cairo_move_to (cr, xc + MARKER_RADIUS * cos(M_PI_2*i) 
			  - (extents.width/2 + extents.x_bearing), // Center text
		       yc - MARKER_RADIUS * sin(M_PI_2*i) 
			  - (extents.height/2 + extents.y_bearing)); // Center text
    cairo_show_text(cr,  axismarkers[dir].mark[i]);
  } 
}

void
draw_2D_plot(int dir, GtkWidget *widget, GdkEventExpose *event)
{
  cairo_t *cr;

  cr = gdk_cairo_create(widget->window);

  draw_2D_axis(cr, dir);
  analyze_diagram(cr, dir);
  cairo_destroy(cr);
}

void
plot_diagram(cairo_t *cr, GArray *points, real ref)
{
  real dA, mag;
  int i, yc, xc;

  dA = 2.0 * M_PI / (points->len - 1);
  xc = PLOT_WIDTH/2;
  yc = PLOT_HEIGHT/2;

  cairo_set_line_width(cr, 1.0);
  cairo_set_source_rgb(cr, PLOT_GRAPH_R, PLOT_GRAPH_G, PLOT_GRAPH_B);

  mag = g_array_index(points, real, 0)/ref;
  cairo_move_to(cr, xc + sum_to_radius(mag) * cos(0.0),
                    yc + sum_to_radius(mag) * sin(0.0));
  for (i = 1; i < points->len; i++) {
    mag = g_array_index(points, real, i)/ref;
    cairo_line_to(cr, xc + sum_to_radius(mag) * cos(i * dA),
                      yc + sum_to_radius(mag) * sin(i * dA));
  }
  cairo_stroke(cr);
}

//-----------------------------------------------------------------------------
//	Antenna analysis

// Calculate total radiated power
real
total_radiated_power(GList *list)
{
  GList *p;
    
  real sum = 0.0;
  for (p = g_list_first(list); p; p = p->next) {
    sum += cabs(((dipole_def *)p)->v);
  }
}

// Analyze the components apported by each of the
// antennas in the array, to one direction (az and el,
// both values in radians)
//
//       xn.x0 + yn.y0 + zn.z0
//  D = ------------------------
//      sqrt(xn^2 + yn^2 + zn^2)
//  Where
//      xn, yn, zn      Normal of the reference plane
//      x0, y0, z0      Coordinates of the point

complex
one_direction(GList *list, real az, real el)
{
  GList *p;
  real dist, xn, yn, zn, denom, exc_abs, exc_arg;
  complex v_sum;
  dipole_def *dip;

  // The reference plane normal
  xn = cos(el) * cos(az);
  yn = sin(el);
  zn = cos(el) * sin(az);

  denom = sqrt(xn*xn + yn*yn + zn*zn);
  v_sum = 0.0;
  // Go over each of the antennas in the list
  for (p = g_list_first(list); p; p = p->next) {
    dip = (dipole_def *)p->data;
    exc_abs = cabs(dip->v);
    exc_arg = carg(dip->v);
    dist = (xn * dip->x + yn * dip->y + zn * dip->z)/denom; // Wavelengths
    dist = dist * 2.0 * M_PI;				    // Radians
    v_sum += exc_abs * cos(exc_arg + dist) +
             exc_abs * sin(exc_arg + dist) * I;
  }
  return v_sum * cos(el);
}

// Analyze the antenna radiation pattern at a certain elevation (elev in radians)
// Set the value of the max voltage detected (*max)
// and return the diagram in a GArray *
GArray *
analyze_at_elevation(real elev, real *max)
{
  GList *list = specs_from_liststore(FALSE);	// Get configuration from the GUI
  GArray *points = g_array_new(FALSE, FALSE, sizeof(real));
  int az_nr;
  real azim, vs, vmax;

  vmax = 0.0;
  for (az_nr = 0; az_nr <= NR_AZ_STEPS; az_nr++) {
    azim = az_nr * AZ_STEP_RAD;
    vs = cabs(one_direction(list, azim, elev));
    if (vs > *max)
      *max = vs;
    g_array_append_vals(points, &vs, 1);
  }
  return points; 
}

// Analyze the antenna radiation pattern at a certain azimuth (azim in radians)
// Set the value of the max voltage detected (*max)
// and return the diagram in a GArray *
// Set new_plot to TRUE to start a new plot. FALSE makes the function
// add points to the previous plot 
GArray *
analyze_at_azimuth(real azim, real *max, gboolean new_plot)
{
  GList *list = specs_from_liststore(FALSE);	// Get configuration from the GUI
  static GArray *points;
  int el_nr;
  real elev, vs;

  if (new_plot) {
    points = g_array_new(FALSE, FALSE, sizeof(real));
  }
  for (el_nr = 0; el_nr <= NR_EL_STEPS; el_nr++) {
    elev = el_nr * EL_STEP_RAD;
    vs = cabs(one_direction(list, azim, elev));
    if (vs > *max)
      *max = vs;
    g_array_append_vals(points, &vs, 1);
  }
  return points; 
}

void
analyze_diagram(cairo_t *cr, int dir)
{
  GArray *xy_points, *xz_points, *yz_points;
  real max = 0.0;
  
  xy_points = analyze_at_azimuth(to_rad(0.0),   &max, TRUE); 
  xy_points = analyze_at_azimuth(to_rad(180.0), &max, FALSE);
  xz_points = analyze_at_elevation(to_rad(0.0), &max); 
  yz_points = analyze_at_azimuth(to_rad(90.0),  &max, TRUE); 
  yz_points = analyze_at_azimuth(to_rad(270.0), &max, FALSE);
  
  switch (dir) {
    case XY_plane: plot_diagram(cr, xy_points, max); 
		   break;
    case YZ_plane: plot_diagram(cr, yz_points, max); 
		   break;
    case ZX_plane: plot_diagram(cr, xz_points, max); 
		   break;
  }
  printf("%.3f\n", max);
}

//-----------------------------------------------------------------------------
//	Signal handlers

gboolean
xy_darea_expose_event(GtkWidget *widget, GdkEventExpose *event)
{
  draw_2D_plot(XY_plane, widget, event);
  return FALSE;
}

gboolean
xz_darea_expose_event(GtkWidget *widget, GdkEventExpose *event)
{
  draw_2D_plot(ZX_plane, widget, event);
  return FALSE;
}

gboolean
yz_darea_expose_event(GtkWidget *widget, GdkEventExpose *event)
{
  draw_2D_plot(YZ_plane, widget, event);
  return FALSE;
}

void
redraw_plot(void)
{
  GtkWidget *nb, *w;
  GdkGLConfig *cfg;
  
  nb = lookup_widget(MainWindow, "plot_notebook");
  switch (gtk_notebook_get_current_page(GTK_NOTEBOOK(nb))) {
    case PLOT_XY:
      w = lookup_widget(MainWindow, "xy_darea");
      break;
    case PLOT_XZ:
      w = lookup_widget(MainWindow, "xz_darea");
      break;
    case PLOT_YZ:
      w = lookup_widget(MainWindow, "yz_darea");
      break;
    case PLOT_3D:
      w = threed_darea; 
      break;
    case PLOT_STEREO:
      w = stereo_darea;
      break;
  }
  gdk_window_invalidate_rect(w->window, NULL, FALSE);
}

void
set_plot_scale(int scale)
{
  pref.plot_scale = scale;
  redraw_plot();
}

////////////////////////////////////////////////////////////////////////////////
//	3D area
////////////////////////////////////////////////////////////////////////////////

void
initialize_gl(void)
{
  stereo_3D_glconfig = gdk_gl_config_new_by_mode(
		GDK_GL_MODE_RGBA | 
		GDK_GL_MODE_DOUBLE |
		GDK_GL_MODE_DEPTH |
		GDK_GL_MODE_ACCUM);
		
  if (stereo_3D_glconfig == NULL) {
    printf("We are missing some capability here!\n");
    _exit(1);
  }
}
 

void
generate_structure(int az_nr, GLfloat verts[], GLshort indices[])
{
  float center, elev, vs,
        az0, az1,
        cos_az0, cos_az1, sin_az0, sin_az1;
  GList    *list = specs_from_liststore(FALSE);	// Get configuration from the GUI
  int   el_nr, i;
  
  center = sscale * sum_to_radius(cabs(one_direction(list, 0.0, -M_PI_2)));

  az0 = az_nr * AZ_STEP_RAD;
  az1 = az0   + AZ_STEP_RAD;
  cos_az0 = cos(az0);    sin_az0 = sin(az0);
  cos_az1 = cos(az1);    sin_az1 = sin(az1);
    
  for (el_nr = 0; 
       el_nr <= NR_EL_STEPS; 
       el_nr++) {
      
    elev = -M_PI/2 + el_nr * EL_STEP_RAD;

    vs = sscale * sum_to_radius(cabs(one_direction(list, az0, elev)));
    verts[el_nr*6 + 0] = vs * cos(elev) * cos_az0; 
    verts[el_nr*6 + 1] = vs * sin(elev);
    verts[el_nr*6 + 2] = vs * cos(elev) * sin_az0;
      
    vs = sscale * sum_to_radius(cabs(one_direction(list, az1, elev)));
    verts[el_nr*6 + 3] = vs * cos(elev) * cos_az1; 
    verts[el_nr*6 + 4] = vs * sin(elev);
    verts[el_nr*6 + 5] = vs * cos(elev) * sin_az1;
      
    for (i = 0; i < 3; i++) {
      indices[el_nr*6 + 0 + i] = el_nr*2 + i + 0;
      indices[el_nr*6 + 3 + i] = el_nr*2 + i + 1;
    }
  }
  for (i = 0; i < 3; i++) {
    indices[el_nr*6 + 0 + i] = i + 0;
    indices[el_nr*6 + 3 + i] = i + 1;
  }
}   


//------------------------------------------------------------------------------
//	OpenGL 3D plot
//------------------------------------------------------------------------------

void
fullscreen_activate(GtkCheckMenuItem *menuitem)
{
  set_fullscreen_mode(gtk_check_menu_item_get_active(menuitem));
}

void
threed_viewport_realize(GtkWidget *widget)
{
  threed_darea = gtk_drawing_area_new();

  gtk_widget_set_gl_capability(threed_darea, stereo_3D_glconfig, 
		NULL,				// Context
		TRUE,				// Direct access
		GDK_GL_RGBA_TYPE);
  
  gtk_widget_set_events(threed_darea,
		GDK_EXPOSURE_MASK | GDK_BUTTON_PRESS_MASK);

  g_signal_connect(G_OBJECT(threed_darea), "realize",
		G_CALLBACK(threed_darea_realize), NULL);
		
  g_signal_connect(G_OBJECT(threed_darea), "configure-event",
		G_CALLBACK(threed_darea_configure_event), NULL);
		
  g_signal_connect(G_OBJECT(threed_darea), "expose-event",
		G_CALLBACK(threed_darea_expose_event), NULL);
		
  gtk_widget_add_events(threed_darea,
		GDK_BUTTON1_MOTION_MASK    | 
		GDK_BUTTON2_MOTION_MASK    | 
		GDK_SCROLL_MASK		   |
		GDK_BUTTON_PRESS_MASK);
		
  g_signal_connect(G_OBJECT(threed_darea), "button-press-event",
		G_CALLBACK(threed_darea_motion_notify_event), NULL);
		
  g_signal_connect(G_OBJECT(threed_darea), "motion-notify-event",
		G_CALLBACK(threed_darea_motion_notify_event), NULL);
		
  g_signal_connect(G_OBJECT(threed_darea), "scroll-event",
		G_CALLBACK(threed_darea_scroll_event), NULL);

  gtk_container_add(GTK_CONTAINER(widget), threed_darea);
  gtk_widget_show(threed_darea);
}

guint
threed_darea_realize(GtkWidget *widget)
{
  if (gdk_gl_drawable_gl_begin(threed_gldrawable, threed_glcontext)) {
    glViewport(0,0, widget->allocation.width, widget->allocation.height);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(-1000,  1000,	     // Left, right
            -1000,  1000,	     // Bottom, top
            -1000,  2500);	     // Near, far
    glMatrixMode(GL_MODELVIEW);
    gdk_gl_drawable_gl_end(threed_gldrawable);
  }
  return TRUE;
}

gboolean
threed_darea_configure_event(GtkWidget *widget)
{
  threed_gldrawable = gtk_widget_get_gl_drawable(widget);
  threed_glcontext  = gtk_widget_get_gl_context(widget);

  if (gdk_gl_drawable_gl_begin(threed_gldrawable, threed_glcontext)) {
    glViewport(0, 0, widget->allocation.width, widget->allocation.height);
    aspect = (real) widget->allocation.width / widget->allocation.height;
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(-1000,  1000,	     // Left, right
            -1000,  1000,	     // Bottom, top
            -1000,  2500);	     // Near, far
    glMatrixMode(GL_MODELVIEW);
    glDisable(GL_DITHER);
    gdk_gl_drawable_gl_end(threed_gldrawable);
  }
  return TRUE;
}

gboolean
threed_darea_expose_event(GtkWidget *widget, GdkEventExpose *event)
{
  if (event->count > 0) 
    return TRUE;

  /* OpenGL functions can be called only if make_current returns true */
  if (gdk_gl_drawable_gl_begin(threed_gldrawable, threed_glcontext))  {
    glClearColor(0, 0.3, 0, 1);
    glClearColor(0, 0,   0, 1);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
    switch (pref.plot_view) {
      case VIEW_3D_LINE:
        glPolygonMode(GL_FRONT_AND_BACK, GL_LINE); break;
      case VIEW_3D_FILL:
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL); break;
    }
    
    //glLightfv(GL_LIGHT0, GL_POSITION, LightPos);
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    glEnable(GL_DEPTH_TEST);
    
    glLoadIdentity();
    glPushMatrix();

    gluLookAt(  0.0, 0.0, 500.0,
                0.0, 0.0,   0.0,
                0.0, 1.0,   0.0);

    glLightfv(GL_LIGHT0, GL_POSITION, LightPos);

    glRotatef( stheta, 1.0, 0.0, 0.0);
    glRotatef(   sphi, 0.0, 1.0, 0.0);
  
    draw_3D_axis();
    draw_3D_plot(AmbientMat, DiffuseMat);
    glPopMatrix();
    
    gdk_gl_drawable_swap_buffers(threed_gldrawable);
  }

  return TRUE;
}

gboolean
threed_darea_motion_notify_event(GtkWidget *widget, GdkEventMotion *event)
{
  GtkAllocation allocation;
  gboolean redraw = FALSE;

  gtk_widget_get_allocation(widget, &allocation);

  if (event->state & GDK_BUTTON1_MASK)
    {
      sphi   += (real)(event->x - beginX) / 4.0;
      stheta -= (real)(beginY - event->y) / 4.0;
      redraw = TRUE;
    }

  if (event->state & GDK_BUTTON2_MASK)
    {
//      sdepth -= ((event->y - beginY)/(allocation.height));
      redraw = TRUE;
    }

  beginX = event->x;
  beginY = event->y;

  if (redraw)  {
    gdk_window_invalidate_rect(gtk_widget_get_window(widget), &allocation, FALSE);
  }
   
  return TRUE;
}

gboolean
threed_darea_button_press_event(GtkWidget *widget, GdkEventButton *event)
{
  GtkWidget *pop;
  
  switch (event->button) {
    case 1:
    case 2:
      beginX = event->x;
      beginY = event->y;
      return TRUE;
  }

  return FALSE;
}

gboolean
threed_darea_scroll_event(GtkWidget *widget, GdkEventScroll *event)
{
  gboolean redraw = FALSE;
  GtkAllocation allocation;

  gtk_widget_get_allocation(widget, &allocation);

  switch (event->direction) {
    case GDK_SCROLL_UP:   sscale *= 1.1; redraw = TRUE; break;
    case GDK_SCROLL_DOWN: sscale /= 1.1; redraw = TRUE; break;
  }

  if (redraw)
    gdk_window_invalidate_rect(gtk_widget_get_window (widget), &allocation, FALSE);

  return FALSE;
}


void
draw_3D_axis(void)
{
  real a, dleft, dup, arrowbegin, arrowend, radius;

  radius     = 1.0 * sscale;
  arrowbegin = sum_to_radius(0.9) * sscale;
  arrowend   = sum_to_radius(1.0) * sscale;

  glBegin(GL_LINES);
  glEnable(GL_LINE_SMOOTH);
  for (a = 0; a < M_PI*2; a += M_PI/12.0) {
    dleft = radius * cos(a);
    dup = radius * sin(a);

    glColor3f(1.0, 0.0, 0.0);		// X: Red
    glVertex3f(0.0,        -dleft,     dup);
    glVertex3f(arrowbegin, -dleft,     dup);
    glVertex3f(arrowbegin, -4.0*dleft, 4.0*dup);
    glVertex3f(arrowend,    0.0,       0.0);

    glColor3f(0.0, 1.0, 0.0);		// Y: Green
    glVertex3f(dleft,      0.0,        dup);
    glVertex3f(dleft,      arrowbegin, dup);
    glVertex3f(4.0*dleft,  arrowbegin, 4.0*dup);
    glVertex3f(0.0,        arrowend,   0.0);

    glColor3f(1.0, 1.0, 0.0);		// Z: Yellow
    glVertex3f(dleft,     -dup,     0.0);
    glVertex3f(dleft,     -dup,     arrowbegin);
    glVertex3f(4.0*dleft, -4.0*dup, arrowbegin);
    glVertex3f(0.0,        0.0,     arrowend);
  }
  glEnd();
}

void
draw_3D_plot(GLfloat Amb[], GLfloat Dif[])
{
  GLfloat  verts  [(NR_EL_STEPS + 1) * 2 * 3];
  GLushort indices[(NR_EL_STEPS + 1) * 2 * 3];
  int az_nr;
  
  for (az_nr = 0; az_nr < NR_AZ_STEPS; az_nr++) {
    generate_structure(az_nr, verts, indices);
    
    glEnableClientState(GL_VERTEX_ARRAY);
    glVertexPointer(3, GL_FLOAT, 0, verts);
    glNormalPointer(GL_FLOAT, 0, verts);
  
    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, Dif);
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, Amb);

    glDrawElements(GL_QUAD_STRIP, 6*(NR_EL_STEPS + 0), GL_UNSIGNED_SHORT, indices);

    glDisableClientState(GL_VERTEX_ARRAY);
  }
}


//------------------------------------------------------------------------------
//      OpenGL stereo plot
//------------------------------------------------------------------------------
	 	
void
stereo_viewport_realize(GtkWidget *widget)
{
  stereo_darea = gtk_drawing_area_new();

  gtk_widget_set_gl_capability(stereo_darea, stereo_3D_glconfig, 
		NULL,			// Context
		TRUE,			// Direct access
		GDK_GL_RGBA_TYPE);
  
  gtk_widget_set_events(stereo_darea,
		GDK_EXPOSURE_MASK          |
		GDK_BUTTON_PRESS_MASK);

  g_signal_connect(G_OBJECT(stereo_darea), "realize",
		G_CALLBACK(stereo_darea_realize), NULL);
		
  g_signal_connect(G_OBJECT(stereo_darea), "configure-event",
		G_CALLBACK(stereo_darea_configure_event), NULL);
		
  g_signal_connect(G_OBJECT(stereo_darea), "expose-event",
		G_CALLBACK(stereo_darea_expose_event), NULL);
		
  gtk_widget_add_events(stereo_darea,
		GDK_BUTTON1_MOTION_MASK    | 
		GDK_BUTTON2_MOTION_MASK    | 
		GDK_SCROLL_MASK		   |
		GDK_BUTTON_PRESS_MASK);
		
  g_signal_connect(G_OBJECT(stereo_darea), "button-press-event",
		G_CALLBACK(stereo_darea_button_press_event), NULL);
		
  g_signal_connect(G_OBJECT(stereo_darea), "motion-notify-event",
		G_CALLBACK(stereo_darea_motion_notify_event), NULL);
		
  g_signal_connect(G_OBJECT(stereo_darea), "scroll-event",
		G_CALLBACK(stereo_darea_scroll_event), NULL);

  gtk_container_add(GTK_CONTAINER(widget), stereo_darea);
  gtk_widget_show(stereo_darea);
}

guint
stereo_darea_realize(GtkWidget *widget)
{
  if (gdk_gl_drawable_gl_begin(stereo_gldrawable, stereo_glcontext))
    {
      glViewport(0,0, widget->allocation.width, widget->allocation.height);
      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      glMatrixMode(GL_MODELVIEW);
      gdk_gl_drawable_gl_end(stereo_gldrawable);
    }
  return TRUE;
}

gboolean
stereo_darea_configure_event(GtkWidget *widget)
{
  stereo_gldrawable = gtk_widget_get_gl_drawable(widget);
  stereo_glcontext  = gtk_widget_get_gl_context(widget);

  if (gdk_gl_drawable_gl_begin(stereo_gldrawable, stereo_glcontext)) {
    glViewport(0,0, widget->allocation.width, widget->allocation.height);
    aspect = (real) widget->allocation.width / widget->allocation.height;
    glDisable(GL_DITHER);
  }
  return TRUE;
}

void
set_frustum(int eye)
{
  real top = nearZ * tan(to_rad(fovy/2));	// sets top of frustum based on fovy and near clipping plane
  real right = aspect * top;			// sets right of frustum based on aspect ratio
  real frustumshift = (IOD/2) * nearZ/screenZ;
	 
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();                                        //reset projection matrix
  switch (eye) {
    case LEFT_EYE:
       glFrustum(-right + frustumshift, 
                  right + frustumshift,
                 -top, top,
                  nearZ, farZ);
       glTranslatef(IOD/2, 0.0, 0.0);       //translate to cancel parallax
       break;
    case RIGHT_EYE:
       glFrustum(-right - frustumshift, 
                  right - frustumshift,
                 -top, top,
                  nearZ, farZ);
       glTranslatef(-IOD/2, 0.0, 0.0);       //translate to cancel parallax
       break;
  }
  glMatrixMode(GL_MODELVIEW);
}

gboolean
stereo_darea_expose_event(GtkWidget *widget, GdkEventExpose *event)
{
  if (event->count > 0) 
    return TRUE;

  /* OpenGL functions can be called only if make_current returns true */
  if (gdk_gl_drawable_gl_begin(stereo_gldrawable, stereo_glcontext))  {
    set_frustum(LEFT_EYE);
    glClearColor(0.1, 0.1, 0.1, 1);
    glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    switch (pref.plot_view) {
      case VIEW_3D_LINE:
        glPolygonMode(GL_FRONT_AND_BACK, GL_LINE); break;
      case VIEW_3D_FILL:
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL); break;
    }
    
    glEnable(GL_DEPTH_TEST);
    
    glLoadIdentity();
    glPushMatrix();

    // LEFT EYE

    glTranslatef(   0.0, 0.0, -camZ);
    glRotatef(   stheta, 1.0, 0.0, 0.0);
    glRotatef(     sphi, 0.0, 1.0, 0.0);
    
    glColorMask(GL_TRUE, GL_FALSE, GL_FALSE, GL_TRUE);
    draw_stereo_axis(AmbientMat_L);
    draw_stereo_plot(AmbientMat_L);
        
    glPopMatrix();
    
    
    // RIGHT EYE
    glClear(GL_DEPTH_BUFFER_BIT);

    set_frustum(RIGHT_EYE);
    glTranslatef(   0.0, 0.0, -camZ);
    glRotatef(   stheta, 1.0, 0.0, 0.0);
    glRotatef(     sphi, 0.0, 1.0, 0.0);

    glColorMask(GL_FALSE, GL_TRUE, GL_TRUE, GL_TRUE);
    draw_stereo_axis(AmbientMat_R);
    draw_stereo_plot(AmbientMat_R);
    

    gdk_gl_drawable_swap_buffers(stereo_gldrawable);
  }

  return TRUE;
}

gboolean
stereo_darea_expose_event_old(GtkWidget *widget, GdkEventExpose *event)
{
  if (event->count > 0) 
    return TRUE;

  /* OpenGL functions can be called only if make_current returns true */
  if (gdk_gl_drawable_gl_begin(stereo_gldrawable, stereo_glcontext))  {
    set_frustum(LEFT_EYE);
    glClearColor(0.1, 0.1, 0.1, 1);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    switch (pref.plot_view) {
      case VIEW_3D_LINE:
        glPolygonMode(GL_FRONT_AND_BACK, GL_LINE); break;
      case VIEW_3D_FILL:
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL); break;
    }
    
    glEnable(GL_DEPTH_TEST);
    
    glLoadIdentity();
    glPushMatrix();

    // LEFT EYE

    glTranslatef(   0.0, 0.0, -camZ);
    glRotatef(   stheta, 1.0, 0.0, 0.0);
    glRotatef(     sphi, 0.0, 1.0, 0.0);
    
    draw_stereo_axis(AmbientMat_L);
    draw_stereo_plot(AmbientMat_L);
    glAccum(GL_LOAD, 1.0); 
        
    glPopMatrix();
    
    
    // RIGHT EYE
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    set_frustum(RIGHT_EYE);
    glTranslatef(   0.0, 0.0, -camZ);
    glRotatef(   stheta, 1.0, 0.0, 0.0);
    glRotatef(     sphi, 0.0, 1.0, 0.0);
    draw_stereo_axis(AmbientMat_R);
    draw_stereo_plot(AmbientMat_R);
    
    glAccum(GL_ACCUM, 1.0);
    glAccum(GL_RETURN, 1.0); 

    gdk_gl_drawable_swap_buffers(stereo_gldrawable);
  }

  return TRUE;
}

gboolean
stereo_darea_motion_notify_event(GtkWidget *widget, GdkEventMotion *event)
{
  GtkAllocation allocation;
  gboolean redraw = FALSE;

  gtk_widget_get_allocation(widget, &allocation);

  if (event->state & GDK_BUTTON1_MASK)
    {
      sphi   += (real)(event->x - beginX) / 4.0;
      stheta -= (real)(beginY - event->y) / 4.0;
      redraw = TRUE;
    }

  if (event->state & GDK_BUTTON2_MASK)
    {
//      sdepth -= ((event->y - beginY)/(allocation.height));
      redraw = TRUE;
    }

  beginX = event->x;
  beginY = event->y;

  if (redraw)  {
    gdk_window_invalidate_rect(gtk_widget_get_window(stereo_darea), &allocation, FALSE);
  }
   
  return TRUE;
}

gboolean
stereo_darea_button_press_event(GtkWidget *widget, GdkEventButton *event)
{
  GtkWidget *pop;
  
  printf("%d\n", event->button);
  switch (event->button) {
    case 1:
      beginX = event->x;
      beginY = event->y;
      return TRUE;
  }
  return FALSE;
}

gboolean
stereo_darea_scroll_event(GtkWidget *widget, GdkEventScroll *event)
{
  gboolean redraw = FALSE;
  GtkAllocation allocation;

  gtk_widget_get_allocation(widget, &allocation);

  switch (event->direction) {
    case GDK_SCROLL_UP:   sscale *= 1.1; redraw = TRUE; break;
    case GDK_SCROLL_DOWN: sscale /= 1.1; redraw = TRUE; break;
  }

  if (redraw)
    gdk_window_invalidate_rect(gtk_widget_get_window(widget), &allocation, FALSE);

  return FALSE;
}


void
draw_stereo_axis(GLfloat col[])
{
  real a, dleft, dup, arrowbegin, arrowend, radius;

  radius     = 1.0 * sscale;
  arrowbegin = sum_to_radius(0.9) * sscale;
  arrowend   = sum_to_radius(1.0) * sscale;

  glBegin(GL_LINES);
  glEnable(GL_LINE_SMOOTH);
  for (a = 0; a < M_PI*2; a += M_PI/12.0) {
    dleft = radius * cos(a);
    dup = radius * sin(a);

    glColor3fv(col);
    glVertex3f(0.0,        -dleft,     dup);		// X
    glVertex3f(arrowbegin, -dleft,     dup);
    glVertex3f(arrowbegin, -4.0*dleft, 4.0*dup);
    glVertex3f(arrowend,    0.0,       0.0);

    glVertex3f(dleft,      0.0,        dup);
    glVertex3f(dleft,      arrowbegin, dup);		// Y
    glVertex3f(4.0*dleft,  arrowbegin, 4.0*dup);
    glVertex3f(0.0,        arrowend,   0.0);

    glVertex3f(dleft,     -dup,     0.0);
    glVertex3f(dleft,     -dup,     arrowbegin);	// Z
    glVertex3f(4.0*dleft, -4.0*dup, arrowbegin);
    glVertex3f(0.0,        0.0,     arrowend);
  }
  glEnd();
}

void
draw_stereo_plot(GLfloat col[])
{
  GLfloat  verts  [(NR_EL_STEPS + 1) * 2 * 3];
  GLushort indices[(NR_EL_STEPS + 1) * 2 * 3];
  int      az_nr;

  for (az_nr = 0; az_nr < NR_AZ_STEPS; az_nr++) {
    generate_structure(az_nr, verts, indices);
    
    glEnableClientState(GL_VERTEX_ARRAY);
    glVertexPointer(3, GL_FLOAT, 0, verts);
    glNormalPointer(GL_FLOAT, 0, verts);
  
    glDrawElements(GL_QUAD_STRIP, 6*(NR_EL_STEPS + 0), GL_UNSIGNED_SHORT, indices);

    glDisableClientState(GL_VERTEX_ARRAY);
  }
}


void
xy1_activate(void)
{
  printf("XY\n");
  sphi = 0.0;
  stheta = 0.0;
  redraw_plot();
}

void
yz1_activate(void)
{
  sphi = 0.0;
  stheta = 90.0;
  redraw_plot();
}

void
zx1_activate(void)
{
  sphi = 90.0;
  stheta = 0.0;
  redraw_plot();
}
